#<profile4xi profile = "11"/>
# -------------------------------------------------------------------------------------------------
#<program4xi program = "parsestudio4xi.py">
#<abstract4xi abstract = "parses studio4xi programs"/>
#<refer4xi reference = "Studio4Xi"/>
#<state4xi status = "ACK4review; ACK4tagging; ACK4moduletesting; ACK4systemtesting; [Rel 7.21]"/>
#<use4xi usage = "
#
# 1) go to Studio4Xi in commmand mode
#       > go3studio4fb <-- Python 3 version of Studio4Xi in Fastblack workstation (fb)
#       Tip: you could use the same go2studio4xi-command in all workstations just varying the content! 
#
# 2) parse2studio4xi
#
# 3) view parsing results in temp folder, for example E:\STUDIO4XI\Python\3.3\temp\parse2studio4xi.txt

# 4) continue
# 4a) if parsing is OK, finish
# 4b) is there are any parsing errors, correct the modules and go to step 2 to parse all again.
#
# Tip: avoid using tags in your comments because parsing may interprete them as program tags!
#"
#/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
__all__ = [ 'parsestudio4xi']


# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
try:                 # system ones
    import os
    import sys
    import signal
except:
    pass
try:                 # threading ones       
    import threading                
    import _thread   
except:
    pass

#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from time4xi import time4xi    
from version4xi import VersionKB4Xi
from version4xi import ROOT4XI
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi Globals                            
# -------------------------------------------------------------------------------------------------
#<globalize4xi globalization = "
global path2parse4studio

global tagindexes4line
global tagminimums4module
global tagstats4module 
global tagstats4system
global tagvalues4attributes
global TagKB4Xi

#"
#"/>

# -------------------------------------------------------------------------------------------------
# Settings
# -------------------------------------------------------------------------------------------------
#<set4xi setting = "


path2parse4studio = VersionKB4Xi['studio']


# -------------------------------------------------------------------------------------------------
# TAGS KNOWLEDGEBASE
# -------------------------------------------------------------------------------------------------

# ==============================================================================================================================================
#   START                          ATTRIBUTES                                                                    SYNONYM             END        
# ==============================================================================================================================================
TagKB4Xi = {
    '<abstract4xi':               {1: 'abstract',                                                                32: '<synopsize4xi',33: '/>'},
    '<acknowledge4xi':            {1: 'acknowledgement', 2: 'tester',                                            32: 'None',         33: '/>'},
    '<define4xi':                 {1: 'definition',                                                              32: 'None',         33: '/>'},
    '<class4xi':                  {1: 'class',           2: 'base',    3: 'child', 4: 'synopsis', 5: 'usage',    32: 'None',         33: '</class4xi>'},
    '<cluster4xi':                {1: 'cluster',                                                                 32: 'None',         33: '/>'},
    '<contextualize4xi':          {1: 'context',                                                                 32: '<refer4xi',    33: '/>'},
    '<define4xi':                 {1: 'definition',                                                              32: 'None',         33: '/>'},
    '<do4xi':                     {1: 'do',                                                                      32: 'None',         33: '/>'},
    '<dragonize4xi':              {1: 'dragonization',                                                           32: 'None',         33: '/>'},
    '<end4xi':                    {1: 'end',                                                                     32: 'None',         33: '/>'},
    '<expose4xi':                 {1: 'expose',                                                                  32: 'None',         33: '/>'},
    '<function4xi':               {1: 'function',                                                                32: 'None',         33: '</function4xi>'},
    '<globalize4xi':              {1: 'globalization',                                                           32: 'None',         33: '/>'},
    '<interface4xi':              {1: 'interface',                                                               32: 'None',         33: '/>'},
    '<link4xi':                   {1: 'module',          2: 'class',   3: 'method', 4: 'function', 5: 'link',    32: 'None',         33: '/>'},
    '<methodize4xi':              {1: 'method',          2: 'usage',                                             32: 'None',         33: '</methodize4xi>'},
    '<note4xi':                   {1: 'note',                                                                    32: 'None',         33: '/>'},
    '<pickle4xi':                 {1: 'pickle',                                                                  32: 'None',         33: '/>'},
    '<process4xi':                {1: 'process',                                                                 32: 'None',         33: '/>'},
    '<profile4xi':                {1: 'profile',                                                                 32: 'None',         33: '/>'},
    '<program4xi':                {1: 'program',                                                                 32: 'None',         33: '</program4xi>'},
    '<protect4xi':                {1: 'protection',                                                              32: 'None',         33: '/>'},
    '<record4xi':                 {1: 'record',                                                                  32: 'None',         33: '/>'},
    '<refer4xi':                  {1: 'reference',                                                               32: '<verify4xi',   33: '/>'},
    '<register4xi':               {1: 'registration',                                                            32: 'None',         33: '/>'},
    '<scope4xi':                  {1: 'galaxy',          2: 'direction', 3: 'distance',                          32: 'None',         33: '/>'},
    '<set4xi':                    {1: 'setting',                                                                 32: 'None',         33: '/>'},
    '<share4xi':                  {1: 'share',                                                                   32: 'None',         33: '/>'},
    '<specialize4xi':             {1: 'speciality',                                                              32: 'None',         33: '/>'},
    '<standardize4xi':            {1: 'standard',                                                                32: 'None',         33: '/>'},
    '<state4xi':                  {1: 'status',                                                                  32: 'None',         33: '/>'},
    '<synopsize4xi':              {1: 'synopsis',                                                                32: '<abstract4xi', 33: '/>'},
    '<translate4xi':              {1: 'translation',                                                             32: 'None',         33: '/>'},
    '<type4xi':                   {1: 'type',                                                                    32: 'None',         33: '/>'},
    '<use4xi':                    {1: 'usage',                                                                   32: 'None',         33: '/>'},
    '<verbalize4xi':              {1: 'None',                                                                    32: 'None',         33: '</verbalize4xi>'},
    '<verify4xi':                 {1: 'verification',    2: 'target',                                            32: 'None',         33: '/>'}
    }

# -------------------------------------------------------------------------------------------------
# INDEXES FOUND FOR EACH TAG IN A PROGRAM LINE, -1 = NOT FOUND
# -------------------------------------------------------------------------------------------------

tagindexes4line = {
    '<abstract4xi':               -1,  # start tags
    '<acknowledge4xi':            -1,
    '<define4xi':                 -1,
    '<class4xi':                  -1,
    '<cluster4xi':                -1,
    '<contextualize4xi':          -1,
    '<define4xi':                 -1,
    '<do4xi':                     -1,
    '<dragonize4xi':              -1,
    '<end4xi':                    -1,
    '<expose4xi':                 -1,
    '<function4xi':               -1,
    '<globalize4xi':              -1,
    '<interface4xi':              -1,
    '<link4xi':                   -1,
    '<methodize4xi':              -1,
    '<note4xi':                   -1,
    '<pickle4xi':                 -1,
    '<process4xi':                -1,
    '<profile4xi':                -1,
    '<program4xi':                -1,
    '<protect4xi':                -1,
    '<record4xi':                 -1,
    '<refer4xi':                  -1,
    '<register4xi':               -1,
    '<scope4xi':                  -1,
    '<set4xi':                    -1,
    '<share4xi':                  -1,
    '<specialize4xi':             -1,
    '<standardize4xi':            -1,
    '<state4xi':                  -1,
    '<synopsize4xi':              -1,
    '<translate4xi':              -1,
    '<type4xi':                   -1,
    '<use4xi':                    -1,
    '<verbalize4xi':              -1,
    '<verify4xi':                 -1,
    '</class4xi>':                -1,  # end tags
    '</define4xi>':               -1,
    '</function4xi>':             -1,
    '</methodize4xi>':            -1,
    '</program4xi>':              -1,
    '</type4xi>':                 -1,
    '</verbalize4xi>':            -1
    }
    
# -------------------------------------------------------------------------------------------------
# TAG STATISTICS FOR SYSTEM
# -------------------------------------------------------------------------------------------------
    
tagstats4system = {
    'zskeleton4xi.py': {'<abstract4xi':           0,  # start tags
                        '<acknowledge4xi':        0,
                        '<class4xi':              0,
                        '<cluster4xi':            0,
                        '<contextualize4xi':      0,
                        '<define4xi':             0,
                        '<do4xi':                 0,
                        '<dragonize4xi':          0,
                        '<end4xi':                0,
                        '<expose4xi':             0,
                        '<function4xi':           0,
                        '<globalize4xi':          0,
                        '<interface4xi':          0,
                        '<link4xi':               0,
                        '<methodize4xi':          0,
                        '<note4xi':               0,
                        '<pickle4xi':             0,
                        '<process4xi':            0,
                        '<profile4xi':            0,
                        '<program4xi':            0,
                        '<protect4xi':            0,
                        '<record4xi':             0,
                        '<refer4xi':              0,
                        '<register4xi':           0,
                        '<scope4xi':              0,
                        '<set4xi':                0,
                        '<share4xi':              0,
                        '<specialize4xi':         0,
                        '<standardize4xi':        0,
                        '<state4xi':              0,
                        '<synopsize4xi':          0,
                        '<translate4xi':          0,
                        '<type4xi':               0,
                        '<use4xi':                0,
                        '<verbalize4xi':          0,
                        '<verify4xi':             0,
                        '</class4xi>':            0,  # end tags
                        '</define4xi>':           0,
                        '</function4xi>':         0,
                        '</methodize4xi>':        0,
                        '</program4xi>':          0,
                        '</type4xi>':             0,
                        '</verbalize4xi>':        0
                        }
      }
      
# -------------------------------------------------------------------------------------------------
# TAGS STATISTICS FOR MODULE
# -------------------------------------------------------------------------------------------------
    
tagstats4module = {
    '<abstract4xi':           0,       # star tags
    '<acknowledge4xi':        0,
    '<class4xi':              0,
    '<cluster4xi':            0,
    '<contextualize4xi':      0,
    '<define4xi':             0,
    '<do4xi':                 0,
    '<dragonize4xi':          0,
    '<end4xi':                0,
    '<expose4xi':             0,
    '<function4xi':           0,
    '<globalize4xi':          0,
    '<interface4xi':          0,
    '<link4xi':               0,
    '<methodize4xi':          0,
    '<note4xi':               0,
    '<pickle4xi':             0,
    '<process4xi':            0,
    '<profile4xi':            0,
    '<program4xi':            0,
    '<protect4xi':            0,
    '<record4xi':             0,
    '<refer4xi':              0,
    '<register4xi':           0,
    '<scope4xi':              0,
    '<set4xi':                0,
    '<share4xi':              0,
    '<specialize4xi':         0,
    '<standardize4xi':        0,
    '<state4xi':              0,
    '<synopsize4xi':          0,
    '<translate4xi':          0,
    '<type4xi':               0,
    '<use4xi':                0,
    '<verbalize4xi':          0,
    '<verify4xi':             0,
    '</class4xi>':            0,       # end tags
    '</define4xi>':           0,
    '</function4xi>':         0,
    '</methodize4xi>':        0,
    '</program4xi>':          0,
    '</type4xi>':             0,
    '</verbalize4xi>':        0
    }
    
      
# -------------------------------------------------------------------------------------------------
# MINIMUM OF TAGS FOR MODULE
# -------------------------------------------------------------------------------------------------
    
tagminimums4module = {
    '<abstract4xi':           1,
    '<acknowledge4xi':        0,
    '<class4xi':              0,
    '<cluster4xi':            0,
    '<contextualize4xi':      0,
    '<define4xi':             0,
    '<do4xi':                 0,
    '<dragonize4xi':          1,
    '<end4xi':                1,
    '<expose4xi':             0,
    '<function4xi':           0,
    '<globalize4xi':          0,
    '<interface4xi':          1,
    '<link4xi':               0,
    '<methodize4xi':          0,
    '<note4xi':               0,
    '<pickle4xi':             0,
    '<process4xi':            0,
    '<profile4xi':            0,
    '<program4xi':            1,
    '<protect4xi':            1,
    '<record4xi':             1,
    '<refer4xi':              1,
    '<register4xi':           0,
    '<scope4xi':              0,
    '<set4xi':                0,
    '<share4xi':              0,
    '<specialize4xi':         0,
    '<standardize4xi':        1,
    '<state4xi':              1,
    '<synopsize4xi':          0,
    '<translate4xi':          0,
    '<type4xi':               0,
    '<use4xi':                0,
    '<verbalize4xi':          0,
    '<verify4xi':             0,
    '</class4xi>':            0,       # end tags
    '</define4xi>':           0,
    '</function4xi>':         0,
    '</methodize4xi>':        0,
    '</program4xi>':          1,
    '</type4xi>':             0,
    '</verbalize4xi>':        0
    }

# -------------------------------------------------------------------------------------------------
# ALLOWED ATTRIBUTE VALUES
# -------------------------------------------------------------------------------------------------
          
tagvalues4attributes = {
    '<profile4xi': {1: '11', 2: '12', 3: '13', 4: '21', 5: '22', 6: '23', 7:'31', 8:'32', 9:'33'} 
    } 

# -------------------------------------------------------------------------------------------------
# ALLOWED ATTRIBUTE VALUES
# -------------------------------------------------------------------------------------------------
#   KEY   FUNCTION4XI CLASS4XI METHODIZE4XI DESCRIPTION       
profiles4tags = {

      '00': {11: 'nack2function',  12: 'nack2special', 21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'no functions nor any classes; for example import4xi.py, verb modules, some xml modules, types'},
      '10': {11: 'ack2function',   12: 'nack2special', 21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'function(s): normal function(s)'},
      '20': {11: 'nack2function',  12: 'nack2special', 21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'Function(s): special function(s)'},
      '30': {11: 'ack2function',   12: 'ack2special',  21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'both noremal function(s) and special Function(s)'},
      '01': {11: 'nack2function',  12: 'nack2special', 21: 'ack2class',   22: 'nack2methods',  23: 'ack2stub',  24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'stub Class'},
      '02': {11: 'nack2function',  12: 'nack2special', 21: 'ack2class',   22: 'ack2methods',   23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'normal Class with methods'},
      '03': {11: 'nack2function',  12: 'nack2special', 21: 'ack2class',   22: 'ack2methods',   23: 'nack2stub', 24: 'ack4embed',  25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'Class with embedded functions !'},
      '11': {11: 'ack2function',   12: 'nack2special', 21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'ack2function',   26: 'nack2special', 27: 'ack2test',  30: 'normal function(s) and normal functions(s); test programs have this profile'},
      '12': {11: 'ack2function',   12: 'nack2special', 21: 'ack2class',   22: 'ack2methods',   23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'normal function(s) and normal Class(es) with methods'},
      '13': {11: 'ack2function',   12: 'nack2special', 21: 'ack2class',   22: 'ack2methods',   23: 'ack2stub',  24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'normal function(s) and both stub Class and normal Class'},
      '21': {11: 'nack2function',  12: 'ack2special',  21: 'ack2class',   22: 'nack2methods',  23: 'ack2stub',  24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'special Function(s) and stub Class'},
      '22': {11: 'nack2function',  12: 'ack2special',  21: 'ack2class',   22: 'ack2methods',   23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'special Function(s) and normal Class with methods'},
      '23': {11: 'nack2function',  12: 'ack2special',  21: 'ack2class',   22: 'ack2methods',   23: 'ack2stub',  24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'special Function(s) and both stub Class and normal Class'},
      '31': {11: 'ack2function',   12: 'ack2special',  21: 'ack2class',   22: 'nack2methods',  23: 'ack2stub',  24: 'nack4embed', 25: 'nack2function',  26: 'nack2special', 27: 'nack2test', 30: 'both normal function(s) and special Function(s) and stub Class'},
      '32': {11: 'ack2function',   12: 'ack2special',  21: 'nack2class',  22: 'nack2methods',  23: 'nack2stub', 24: 'nack4embed', 25: 'nack2function',  26: 'ack2special',  27: 'nack2test', 30: 'both normal function(s) and special Function(s) and in body special Functions(s)'},
      '33': {11: 'ack2function',   12: 'ack2special',  21: 'ack2class',   22: 'ack2methods',   23: 'ack2stub',  24: 'ack4embed',  25: 'ack2function',   26: 'ack2special',  27: 'ack2test',  30: 'ALL' }
 
 }
            
#"
#"/>


# -------------------------------------------------------------------------------------------------
#<function4xi function = "parsestudio4xi" usage = "
#   parsing a dranonized source file
#"> 
# -------------------------------------------------------------------------------------------------
def parsestudio4xi(fullName):
    global path2parse4studio
    global tagindexes4line
    global tagminimums4module
    global tagstats4module 
    global tagstats4system
    global tagvalues4attributes
    global TagKB4Xi
    
    # -------------------------------------------------------------------------------------------------    
    # STEP 1
    # -------------------------------------------------------------------------------------------------         
            
    name2xi=os.path.splitext(fullName)
    
    print ('STARTED ', time4xi())
    print ('PASSED FULLNAME: ' + fullName)    
    print ('PASSED NAME2XI[0]: ' + name2xi[0])
    print ('PASSED NAME2XI[1]: ' + name2xi[1])

           
    # -------------------------------------------------------------------------------------------------    
    # STEP 2
    # -------------------------------------------------------------------------------------------------
    text2find = '4xi.py'
    index2dragon = str.find(fullName, text2find)  
    if index2dragon == -1:
        ack2dragon = False           # non dragonixed module including *2xi.py
        print ('(STEP 2A/33) OF PARSE4XI - DETERMINED TO BE NON DRAGONIZED FILE: ', fullName, ' INDEX2DRAGON: ', index2dragon)
                 
    else:    
        ack2dragon = True            # dragonixed module either *4xi.py or 4xi.pyc 
        print ('(STEP 2B/33) OF PARSE4XI - DRAGONIZED SOURCE OR COMPILED ONE: ', fullName, ' INDEX2DRAGON: ', index2dragon)

        
    # -------------------------------------------------------------------------------------------------    
    # STEP 3
    # -------------------------------------------------------------------------------------------------        
    if ((name2xi[1] == '.py') and (ack2dragon is True)):
        # ---------------------------------------------------------------------------------------------
        # STEP 3: DRAGONIZED PYTHON SOURCE FILE BEING PARSED
        # ---------------------------------------------------------------------------------------------    
        file2xi = name2xi[0]+'.py'
        print ('(STEP 3A/33) OF PARSE4XI - PARSING DRAGONIZED PYTHON SOURCE FILE: ', file2xi, ' ', time4xi())

                            
        # ---------------------------------------------------------------------------------------------
        # INITIALIZE STATISTICS FOR THIS MODULE
        # ---------------------------------------------------------------------------------------------                                     
        tagstats4module = {
            '<abstract4xi':           0,       # start tags
            '<acknowledge4xi':        0,
            '<class4xi':              0,
            '<cluster4xi':            0,
            '<contextualize4xi':      0,
            '<define4xi':             0,
            '<do4xi':                 0,
            '<dragonize4xi':          0,
            '<end4xi':                0,
            '<expose4xi':             0,
            '<function4xi':           0,
            '<globalize4xi':          0,
            '<interface4xi':          0,
            '<link4xi':               0,
            '<methodize4xi':          0,
            '<note4xi':               0,
            '<pickle4xi':             0,
            '<process4xi':            0,
            '<profile4xi':            0,
            '<program4xi':            0,
            '<protect4xi':            0,
            '<record4xi':             0,
            '<refer4xi':              0,
            '<register4xi':           0,
            '<scope4xi':              0,
            '<set4xi':                0,
            '<share4xi':              0,
            '<specialize4xi':         0,
            '<standardize4xi':        0,
            '<state4xi':              0,
            '<synopsize4xi':          0,
            '<translate4xi':          0,
            '<type4xi':               0,
            '<use4xi':                0,
            '<verbalize4xi':          0,
            '<verify4xi':             0,
            '</class4xi>':            0,       # end tags
            '</define4xi>':           0,
            '</function4xi>':         0,
            '</methodize4xi>':        0,
            '</program4xi>':          0,
            '</type4xi>':             0,
            '</verbalize4xi>':        0
            }            
            
        # ---------------------------------------------------------------------------------------------
        # INITIALIZE SYSTEM STATISTICS DUE TO THIS MODULE
        # ---------------------------------------------------------------------------------------------                
        tagstats4system[file2xi] = {
                        '<abstract4xi':           0,  # start tags
                        '<acknowledge4xi':        0,
                        '<class4xi':              0,
                        '<cluster4xi':            0,
                        '<contextualize4xi':      0,
                        '<define4xi':             0,
                        '<do4xi':                 0,
                        '<dragonize4xi':          0,
                        '<end4xi':                0,
                        '<expose4xi':             0,
                        '<function4xi':           0,
                        '<globalize4xi':          0,
                        '<interface4xi':          0,
                        '<link4xi':               0,
                        '<methodize4xi':          0,
                        '<note4xi':               0,
                        '<pickle4xi':             0,
                        '<process4xi':            0,
                        '<profile4xi':            0,
                        '<program4xi':            0,
                        '<protect4xi':            0,
                        '<record4xi':             0,
                        '<refer4xi':              0,
                        '<register4xi':           0,
                        '<scope4xi':              0,
                        '<set4xi':                0,
                        '<share4xi':              0,
                        '<specialize4xi':         0,
                        '<standardize4xi':        0,
                        '<state4xi':              0,
                        '<synopsize4xi':          0,
                        '<translate4xi':          0,
                        '<type4xi':               0,
                        '<use4xi':                0,
                        '<verbalize4xi':          0,
                        '<verify4xi':             0,
                        '</class4xi>':            0,  # end tags
                        '</define4xi>':           0,
                        '</function4xi>':         0,
                        '</methodize4xi>':        0,
                        '</program4xi>':          0,
                        '</type4xi>':             0,
                        '</verbalize4xi>':        0
                        }                     
            
        # ---------------------------------------------------------------------------------------------
        # OPEN DRAGONIZED PYTHON SOURCE FILE
        # ---------------------------------------------------------------------------------------------          
        input_file2xi = open(file2xi)

                  
        # ---------------------------------------------------------------------------------------------
        # STEP 4 - ALL LINES IN A DRAGONIZED SOURCE FILE
        # ---------------------------------------------------------------------------------------------
        while 1:
   
            # -----------------------------------------------------------------------------------------    
            # STEP 5 - LINE
            # -----------------------------------------------------------------------------------------     
            line2xi = input_file2xi.readline() 
            if not line2xi:
                # -------------------------------------------------------------------------------------    
                # STEP 32 - ALL LINES PARSED - CLOSE FILE
                # -------------------------------------------------------------------------------------     
                print ('(STEP 32A/33) OF PARSE4XI - PARSED DRAGONIZED PYTHON SOURCE FILE: ', file2xi, ' ', time4xi())

                try:
                    input_file2xi.close()
                except:
                    print ('(STEP 32B/33) OF PARSE4XI *** ERROR *** FAILED TO CLOSE DRAGONIZED PYTHON SOURCE FILE: ', file2xi, ' ', time4xi())    
                break                                                          # all lines read, finish
            else:
                # -------------------------------------------------------------------------------------
                # STEP 7 - CHECK ALL TAGS IN PROGRAM LINE
                # -------------------------------------------------------------------------------------             
                keys2indexes = tagindexes4line.keys()
                for key2index in sorted(keys2indexes):
                    tagindexes4line[key2index] = str.find(line2xi, key2index)  # <program4xi ... </program4xi>     
                                                                               # -1 = not found
                # -------------------------------------------------------------------------------------
                # STEP 8 - UPDATE DRAGONIZED MODULE STATISTICS
                # -------------------------------------------------------------------------------------           
                keys2tags = tagstats4module.keys()
                for key2tag in sorted(keys2tags):
                    if tagindexes4line[key2tag] == -1:
                        pass
                    else:
                        try:     
                            tagstats4module[key2tag] = tagstats4module[key2tag] + 1                  
                        except:
                            print ('(STEP 8/33) OF PARSE4XI *** ERROR *** FAILED TO UPDATE MODULE STATISTICS FOR DRAGONIZED FILE2XI: ', file2xi, ' ', time4xi())
                            
                # -------------------------------------------------------------------------------------
                # STEP 9 - UPDATE SYSTEM STATISTICS DUE TO THIS MODULE READING MODULE STATISTICS
                # -------------------------------------------------------------------------------------           
                keys2tags = tagstats4module.keys()
                for key2tag in sorted(keys2tags):
                    if tagindexes4line[key2tag] == -1:
                        pass
                    else:
                        try:
                            tagstats4system[file2xi] [key2tag] =  tagstats4system[file2xi] [key2tag] + 1     
                        except:
                            print ('(STEP 9/33) OF PARSE4XI *** ERROR *** FAILED TO UPDATE SYSTEM STATISTICS FOR DRAGONIZED FILE2XI: ', file2xi, ' ', time4xi())
                                             
                # -------------------------------------------------------------------------------------
                # EOF CHECKING TAGS IN PROGRAM LINE
                # ------------------------------------------------------------------------------------- 
                                                     
            # -----------------------------------------------------------------------------------------    
            # END OF WHILE CHECKS
            # -----------------------------------------------------------------------------------------     

        # ---------------------------------------------------------------------------------------------    
        # EOF WHILE
        # ---------------------------------------------------------------------------------------------     
        print ('(STEP 10A/33) OF PARSE4XI - EOF WHILE FOR DRAGONIZED FILE: ', file2xi, ' ', time4xi())

                
    # -------------------------------------------------------------------------------------------------    
    # STEP 2 - SKIPPING NON DRAGONIZED MODULE
    # -------------------------------------------------------------------------------------------------    
    else:
        print ('(STEP 2B/33) OF PARSE4XI - SKIPPED PARSING - NOT DRAGONIZED AND/OR NOT SOURCE FILE: ', fullName, ' ', time4xi())         

                                            
    # ---------------------------------------------------------------------------------------------
    # END OF PARSING
    # ---------------------------------------------------------------------------------------------
    print(33*("="))
    print ('(STEP 33/33) OF PARSE4XI - RETURNING ', time4xi())
    print(33*("="))
               
#</function4xi>    

# =================================================================================================
#<class4xi class = "DirectoryWalker4Xi" base = "None" child = "None" 
#   synopsis = "a forward iterator that traverses a directory tree"
#   usage = "traversing a directory tree"
#>
# =================================================================================================
class DirectoryWalker4Xi:

    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "init" usage = "initializes DirectoryWalker4Xi"> 
    # ---------------------------------------------------------------------------------------------
    def __init__(self, directory):
        self.stack = [directory]
        self.files = []
        self.index = 0
    #</methodize4xi>
            
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "getitem" usage = "gets item"> 
    # ---------------------------------------------------------------------------------------------
    def __getitem__(self, index):
        while 1:
            try:
                file = self.files[self.index]
                self.index = self.index + 1
            except IndexError:
                # pop next directory from stack
                self.directory = self.stack.pop()
                self.files = os.listdir(self.directory)
                self.index = 0
            else:
                # got a filename
                fullname = os.path.join(self.directory, file)
                if os.path.isdir(fullname) and not os.path.islink(fullname):
                    self.stack.append(fullname)
                return fullname
    #</methodize4xi>
    
#</class4xi>        
    
# -------------------------------------------------------------------------------------------------
##<function4xi function = "main" usage = "
#   main function of parsestudio4xi.py
#">                           
# -------------------------------------------------------------------------------------------------
def main():
    global path2parse4studio
    global tagindexes4line
    global tagminimums4module
    global tagstats4module 
    global tagstats4system
    global tagvalues4attributes
    global TagKB4Xi
    
    # ---------------------------------------------------------------------------------------------
    #
    # ---------------------------------------------------------------------------------------------    
    tagstats4system = {} 
    
    count2errors = 0
    count2failures = 0
      
    # ---------------------------------------------------------------------------------------------
    # (STEP 1/33)
    # ---------------------------------------------------------------------------------------------        
    keys2tags = TagKB4Xi.keys()


    print ('(STEP 1B/33) - TAGS TO BE PARSED: ', time4xi())  
    # only start tags - end tags are defined by start tags
               
    for key2tag in sorted(keys2tags):
        print (key2tag, '/>')     
                # <abstract4xi />
                # <acknowledge4xi />
                # <class4xi />
                # <cluster4xi />
                # <contextualize4xi />
                # <define4xi />
                # <do4xi />
                # <dragonize4xi />
                # <end4xi />
                # <expose4xi />
                # <function4xi />
                # <globalize4xi />
                # <interface4xi />
                # <link4xi />
                # <methodize4xi />
                # <note4xi />
                # <pickle4xi />
                # <process4xi />
                # <profile4xi />
                # <program4xi />
                # <protect4xi />
                # <record4xi />
                # <refer4xi />
                # <register4xi />
                # <scope4xi />
                # <set4xi />
                # <share4xi />
                # <specialize4xi />
                # <standardize4xi />
                # <state4xi />
                # <synopsize4xi />
                # <translate4xi />
                # <type4xi />
                # <use4xi />
                # <verbalize4xi />
                # <verify4xi /> 
    
    # ---------------------------------------------------------------------------------------------
    # (STEP 2/33)
    # ---------------------------------------------------------------------------------------------                  
    print('=' *70)
    print("(STEP 2/33) - EXPLORING ", path2parse4studio)
    print('=' *70)
        
    # ---------------------------------------------------------------------------------------------
    # (STEP 3/33) - WALKING
    # ---------------------------------------------------------------------------------------------     
    for file2xi in DirectoryWalker4Xi(path2parse4studio):
    
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------      
        if os.path.isdir(file2xi):
            print ("(STEP 3A/33) - EXPLORING DIRECTORY: " + file2xi)

        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------                       
        if os.path.isfile(file2xi):
            print ("(STEP 3B/33) - EXPLORING FILE: " + file2xi)

            # -------------------------------------------------------------------------------------
            # PARSE
            # -------------------------------------------------------------------------------------                     
            parsestudio4xi(file2xi)
            
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------             
        
    # ---------------------------------------------------------------------------------------------
    # (STEP 4/33) - SYSTEM LEVEL PARSING STATISTICS
    # ---------------------------------------------------------------------------------------------               
    keys2modules = tagstats4system.keys()
    keys2tags = TagKB4Xi.keys()
    
    print ("(STEP 4A/33) - SYSTEM LEVEL PARSING STATISTICS ", time4xi()) 
     
    # --------------------------------------------------------------------------------------------------------------
    # MODULES
    # --------------------------------------------------------------------------------------------------------------      
      
    for key2module in sorted(keys2modules):
            
        # -------------------------------------------------------------------------------------------------------
        # (STEP 5/33) - MODULE LEVEL PARSING STATISTICS
        # -------------------------------------------------------------------------------------------------------        
        print('=' *70)
        print ("(STEP 5A/33) - MODULE LEVEL PARSING STATISTICS FOR ", key2module, "  ", time4xi())
        print('=' *70)        
        # -------------------------------------------------------------------------------------------------------
        # KEYS
        # -------------------------------------------------------------------------------------------------------                       
        for key2tag in sorted(keys2tags):
            if key2tag[1] == '/':
                # --------------------------------------------------------------------------------------------
                # SKIP PARSING END TAG
                # --------------------------------------------------------------------------------------------   
                pass
            else:
                # ---------------------------------------------------------------------------------------------
                # (STEP 6/33) - TAG PARSING
                # ---------------------------------------------------------------------------------------------        
                ack2tag = True  # default is ok for tags
                print ("(STEP 6A/33) - PARSING TAG: ", key2tag, ': ', tagstats4system[key2module] [key2tag])  
                # (STEP 6A/33) - PARSING TAG:  <function4xi :  1
                 
                # ---------------------------------------------------------------------------------------------
                # TRY TAG PARSING
                # ---------------------------------------------------------------------------------------------             
                try:
                    # -----------------------------------------------------------------------------------------
                    # 
                    # -----------------------------------------------------------------------------------------     
                    if tagstats4system[key2module] [key2tag] < tagminimums4module [key2tag]:
                        # -------------------------------------------------------------------------------------
                        # TAG EXISTS LESS THAN REQUIRED 
                        # -------------------------------------------------------------------------------------                           
                        try:
                            # ---------------------------------------------------------------------------------
                            # (STEP 7/33) - CHECK IF SYNONYM TAG CAN HELP 
                            # ---------------------------------------------------------------------------------
                            key2synonym = TagKB4Xi[key2tag] [32]
                            if key2synonym == 'None':
                                # -----------------------------------------------------------------------------
                                # MISSING TAG HAS NO SYNONYM  
                                # -----------------------------------------------------------------------------                            
                                ack2tag = False
                                count2errors = count2errors + 1  
                                print ("(STEP 7A/33) *** ERROR IN MODULE: ", key2module, " MISSING TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag], " < ", tagminimums4module [key2tag]) 
                                # ---------------------------------------------------------------------------------
                                # 
                                # ---------------------------------------------------------------------------------                                              
                            else:
                                # -----------------------------------------------------------------------------
                                # MISSING TAG HAS SYNONYM TAG - CHECK ITS OCCURRENCY 
                                # -----------------------------------------------------------------------------
                                print ("(STEP 7B/33) - CHECKING IN MODULE SYNONUM TAG: ", key2synonym, " QTY: ", tagstats4system[key2module] [key2synonym], " MINUM: ", tagminimums4module [key2synonym])
                                
                                # -----------------------------------------------------------------------------
                                # 
                                # -----------------------------------------------------------------------------          
                                if ((tagstats4system[key2module] [key2synonym] < tagminimums4module [key2synonym]) or (tagstats4system[key2module] [key2synonym] == 0)):
                                    # -------------------------------------------------------------------------
                                    # SYNONYM TAG DID NOT OCCUR IN MODULE  
                                    # -------------------------------------------------------------------------
                                    ack2tag = False
                                    count2errors = count2errors + 1 
                                    print ("(STEP 7C/33) *** ERROR IN MODULE: ", key2module, " MISSING TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag], " < ", tagminimums4module [key2tag])
                                    print ("(STEP 7D/33) *** ERROR IN MODULE: ", key2module, " SYNONYM TAG: ", key2synonym, " QTY: ", tagstats4system[key2module] [key2synonym]) 
                                    # ..//..//..//..//studio4xi//python//3.1//src//load2eight4xi.py  MISSING TAG:  <abstract4xi  QTY:  0  <  1
                                    # ..//..//..//..//studio4xi//python//3.1//src//load2eight4xi.py  SYNONYM TAG:  <synopsize4xi  QTY:  0
                                    # -----------------------------------------------------------------------------
                                    # 
                                    # -----------------------------------------------------------------------------                                             
                                else:
                                    # -------------------------------------------------------------------------
                                    # SYNONYM TAG EXISTS  
                                    # -------------------------------------------------------------------------
                                    ack2tag = True
                                    print ("(STEP 7E/33) - ACCEPTED TAG: ", key2tag, " SYNONYM: ", key2synonym, " QTY-SYNONYM: ", tagstats4system[key2module] [key2synonym])
                                    
                                    # -----------------------------------------------------------------------------
                                    # EOF SYNONYM CHECK 
                                    # -----------------------------------------------------------------------------                                                    
                            # -------------------------------------------------------------------------------------
                            # 
                            # -------------------------------------------------------------------------------------  
                        except:
                            print ("(STEP 7F/33) *** ERROR *** FAILED IN SYNONYM HANDLING ***") 
                            count2failures = count2failures + 1                                                                                                             
                    else:
                        # -----------------------------------------------------------------------------------------
                        # TRYING TAG EXISTS AS REQUIRED OR TAG NOT PRESENT AT ALL  
                        # -----------------------------------------------------------------------------------------                      
                        try:
                            # -------------------------------------------------------------------------------------
                            # TAG EXISTS AS REQUIRED OR TAG NOT PRESENT AT ALL   
                            # -------------------------------------------------------------------------------------                                  
                            if tagstats4system[key2module] [key2tag] > 0:
                                # ---------------------------------------------------------------------------------
                                # TAG FOUND ON MODULE   
                                # ---------------------------------------------------------------------------------           
                                try:
                                    # -----------------------------------------------------------------------------
                                    # CHECK POSSIBLE END TAG FOR THIS TAG 
                                    # -----------------------------------------------------------------------------                                
                                    try:
                                        key2end = TagKB4Xi[key2tag] [33]  
                                        
                                        try:
                                            if key2end == '/>':
                                                # ---------------------------------------------------------------------
                                                # NO SPECIFIC END TAG FOR THIS TAG - NOTHING TO CHECK  
                                                # ---------------------------------------------------------------------  
                                                pass
                                            else:
                                                # -------------------------------------------------------------------------
                                                # CHECK WHETHER NUMBER OF START TAGS MATCHES TO THAT OF END TAGS  
                                                # -------------------------------------------------------------------------  
                                                try:
                                                    if tagstats4system[key2module] [key2tag] == tagstats4system[key2module] [key2end]:
                                                        # ---------------------------------------------------------------------
                                                        # NUMBER OF START TAGS MATCHES TO THAT OF END TAGS  
                                                        # ---------------------------------------------------------------------  
                                                        print ("(STEP 8A/33) - MATCHING TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag], " EQ ", tagstats4system[key2module] [key2end], " = QTY OF END_TAG: ", key2end)
                                                        # (STEP 8A/33) - MATCHING TAG:  <function4xi  QTY:  1  EQ  1  = QTY OF END_TAG:  </function4xi>
                                                        ack2tag = True                                                                 
                                                    else:
                                                        # ---------------------------------------------------------------------
                                                        # MISMATCH BETWEEN NUMBER OF START AND NUMBER OF END TAGS  
                                                        # ---------------------------------------------------------------------                                      
                                                        if ((key2module == "..//..//..//..//studio4xi//python//3.1//src//parse4xi.py") or (key2module == "..//..//..//..//studio4xi//python//3.1//src//parsestudio4xi.py")):
                                                            pass
                                                        else:    
                                                            ack2tag = False
                                                            count2errors = count2errors + 1
                                                            print ("(STEP 8B/33) *** ERROR IN MODULE: ", key2module, " TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag], " N.E. ", tagstats4system[key2module] [key2end], " = QTY OF END_TAG: ", key2end)
                                            
                                                        # -------------------------------------------------------------------------
                                                        # EOF CHECK WHETHER NUMBER OF START TAGS MATCHES TO THAT OF END TAGS  
                                                        # -------------------------------------------------------------------------
                                                except:
                                                    # -----------------------------------------------------------------------------
                                                    # EXCEPTION TAG EXISTS & AS REQUIRED  
                                                    # -----------------------------------------------------------------------------                                        
                                                    print ("(STEP 9A/33) *** ERROR *** FAILED ***") 
                                                    count2failures = count2failures + 1
                                                    try:
                                                        print ("(STEP 9B/33) KEY2MODULE: ", key2module, " KEY2TAG: ", key2tag, " KEY2END: ", key2end)
                                                        # _bootstrap4xi.py  KEY2TAG:  <function4xi  KEY2END:  </function4xi> 
                                                    except:
                                                        pass
                                                                                                                               
                                            # -----------------------------------------------------------------------------
                                            # EOF TAG EXISTS & AS REQUIRED  
                                            # -----------------------------------------------------------------------------
                                        except:
                                            # -----------------------------------------------------------------------------
                                            # EXCEPTION TAG EXISTS & AS REQUIRED  
                                            # -----------------------------------------------------------------------------                                        
                                            print ("(STEP 10A/33) *** ERROR *** FAILED ***") 
                                            count2failures = count2failures + 1 
                                            try:
                                                print ("(STEP 10B/33) KEY2MODULE: ", key2module, " KEY2TAG: ", key2tag, " KEY2END: ", key2end) 
                                            except:
                                                pass                                                                                          
                                    except:
                                        # -----------------------------------------------------------------------------
                                        # EXCEPTION TAG EXISTS & AS REQUIRED  
                                        # -----------------------------------------------------------------------------                                        
                                        print ("(STEP 11A/33) *** ERROR *** FAILED ***") 
                                        count2failures = count2failures + 1 
                                        try:
                                            print ("(STEP 11B/33) KEY2MODULE: ", key2module, " KEY2TAG: ", key2tag, " KEY2END: ", key2end) 
                                        except:
                                            pass                                                                                    
                                except:
                                    # -----------------------------------------------------------------------------
                                    # EXCEPTION TAG EXISTS & AS REQUIRED  
                                    # -----------------------------------------------------------------------------                                        
                                    print ("(STEP 12A/33) *** ERROR *** FAILED ***") 
                                    count2failures = count2failures + 1 
                                    try:
                                        print ("(STEP 12B/33) KEY2MODULE: ", key2module, " KEY2TAG: ", key2tag, " KEY2END: ", key2end) 
                                    except:
                                        pass                                          
                                # ---------------------------------------------------------------------------------
                                # EOF TAG PRESENT   
                                # ---------------------------------------------------------------------------------                                                                                                        
                            else:        
                                # ---------------------------------------------------------------------------------
                                # TAG NOT PRESENT AT ALL  
                                # ---------------------------------------------------------------------------------                                               
                                pass
                            # -------------------------------------------------------------------------------------
                            # EOF TAG EXISTS AS REQUIRED OR TAG NOT PRESENT AT ALL   
                            # -------------------------------------------------------------------------------------                                          
                        except:
                            # -------------------------------------------------------------------------------------
                            # EXCEPTION TAG EXISTS AS REQUIRED OR TAG NOT PRESENT AT ALL   
                            # -------------------------------------------------------------------------------------                                   
                            print ("(STEP 13/33) *** ERROR *** FAILED ***") 
                            count2failures = count2failures + 1         
                            
                        # ---------------------------------------------------------------------------------
                        # EOF TRYING TAG EXISTS AS REQUIRED OR TAG NOT PRESENT AT ALL 
                        # ---------------------------------------------------------------------------------                                       
                        if key2tag[1] == '/':
                            pass
                        else:
                            if ack2tag is True:    
                                print ("(STEP 14A/33) - IN MODULE: ", key2module, " ACCEPTED TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag]) 
                            else:
                                print ("(STEP 14B/33) *** ERROR WITH TAG: ", key2tag, " QTY: ", tagstats4system[key2module] [key2tag], " IN MODULE: ", key2module) 
                                                
                        # ----------------------------------------------------------------------------------
                        # EOF TAG PARSING
                        # ----------------------------------------------------------------------------------
                                                                   
                except:
                    # ------------------------------------------------------------------------------------------
                    # FAILED TAG PARSING
                    # ------------------------------------------------------------------------------------------                        
                    print ("(STEP 15/33) *** ERROR *** FAILED IN TAG LEVEL PARSING STATISTICS ", time4xi()) 
                    count2failures = count2failures + 1         

        # ----------------------------------------------------------------------------------------------------------
        # EOF KEYS  
        # ---------------------------------------------------------------------------------------------------------- 
                                         
    # --------------------------------------------------------------------------------------------------------------
    # EOF MODULES
    # --------------------------------------------------------------------------------------------------------------      

                                        
    # --------------------------------------------------------------------------------------------------------------
    # (STEP 33/33)
    # --------------------------------------------------------------------------------------------------------------              
     
    print('=' *140)
    # ----------------------------------------------------------------------------------------------------------
    # 
    # ----------------------------------------------------------------------------------------------------------                 
    try:
        print("(STEP 33A/33) PARSED TOTALLY ", len(TagKB4Xi), " TAGS IN : ", len(tagstats4system), " MODULES ", time4xi())
        # 36  TAGS IN :  52  MODULES  13:35:12 05/27/14 FLE Daylight Time
    except:
        pass
        
    if count2errors == 0:
        print("(STEP 33B/33) + OK PARSING ", path2parse4studio, " MODULES ", time4xi())
        # ..//..//..//..//studio4xi//python//3.3//src//  MODULES  13:35:12 05/27/14 FLE Daylight Time
              
    else:
        print("(STEP 33B/33)  *** ", count2errors, " ERRORS IN PARSING ", path2parse4studio, " MODULES ", time4xi())
              
    # ----------------------------------------------------------------------------------------------------------
    # 
    # ----------------------------------------------------------------------------------------------------------                 
    if count2failures == 0:
        print("(STEP 33C/33) + OK EXPLORING FOLDER ", path2parse4studio, " ", time4xi())
        # ..//..//..//..//studio4xi//python//3.3//src//   13:35:12 05/27/14 FLE Daylight Time
    else:
        print("(STEP 33C/33)  *** ", count2failures, " FAILURES EXPLORING FOLDER ", path2parse4studio, " ", time4xi())
          
    print('=' *140)

#</function4xi>    
    
        
# -------------------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    main()        

#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   parsestudio4xi.txt
# ============================================================================================================================================
# PARSED TOTALLY  36  TAGS IN :  52  MODULES  05:11:37 05/27/14 Pacific Daylight Time
# + OK PARSING  ..//..//..//..//studio4xi//python//3.3//src//  MODULES  05:11:37 05/27/14 Pacific Daylight Time
# + OK EXPLORING FOLDER  ..//..//..//..//studio4xi//python//3.3//src//   05:11:37 05/27/14 Pacific Daylight Time
# ============================================================================================================================================
# added tags as comments to get this module parsed by itself!
#</function4xi>
#</function4xi>
#</class4xi>
#</methodize4xi>
#</program4xi>
#</verbalize4xi>
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>

#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of parsestudio4xi.py                            
# ------------------------------------------------------------------------------------------------- 
#"
#/>
